package SZXX

  1. Overview
  2. Docs

Module Expert.SSTSource

Sourcetype t
Sourceval filter_path : Base.string Base.list

Advanced: for use with SZXX.Xml.SAX.Expert.Stream.folder

Sourceval zip_entry_filename : Base.string

Advanced: for use with SZXX.Zip.stream_files

Sourceval from_feed : Feed.t -> t

Extract the SST from an XLSX document. This function will stop reading from the Feed as soon as it has retrieved the SST.

Sourceval from_file : _ Eio.File.ro -> t

Extract the SST from an XLSX file. This function does not advance the file cursor. It jumps around the file to only extract the SST while reading as few bytes as necessary.

Sourceval resolve_sst_index : t -> sst_index:Base.string -> Base.string Base.option

Resolve a single reference into the Shared Strings Table.