package SZXX
Page
Library
Module
Module type
Parameter
Class
Class type
Module Xml.DOMSource
Basic XML types and accessor functions
type element = {tag : Base.string;attrs : attr_list;text : Base.string;children : element Base.list;
}Convenience function to access attributes by name
Convenience function to check whether an element has attribute xml:space="preserve"
SZXX.Xml.DOM.unescape "Fast & Furious 🏎️" returns "Fast & Furious 🏎️"
el |> dot "row" returns the first immediate <row> child of element el
el |> dot "row" returns the text of the first immediate <row> child of element el
el |> filter_map "row" ~f returns all filtered f <row> children of element el
el |> at 3 returns the nth (0-based indexing) immediate child of element el.
Same as at but takes a string argument. This allows at_s to be used in get together with dot and dot_text
get el [dot "abc"; dot "def"] is equivalent to el |> dot "abc" |> Option.bind ~f:(dot "def") Convenience function to chain multiple dot and at calls to access nested elements.