package SZXX
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Module Zip.DataSource
type 'a parser_state = | Success of 'a| Failed of {error : Base.string;unconsumed : Base.string;(*This string contains the start of the data not consumed by the parser. More data might have been present but it was not kept to prevent unbounded memory usage.
*)
}| Terminated_early of {unconsumed : Base.string;(*This string contains the start of the data not consumed by the parser. More data might have been present but it was not kept to prevent unbounded memory usage.
*)
}| Incomplete
val compare_parser_state :
('a -> 'a -> Base.int) ->
'a parser_state ->
'a parser_state ->
Base.intSource
val equal_parser_state :
('a -> 'a -> Base.bool) ->
'a parser_state ->
'a parser_state ->
Base.boolSource
type 'a t = | Skip| Fast_skip| String of Base.string| Bigstring of Bigstringaf.t| Fold_string of 'a| Fold_bigstring of 'a| Parse of 'a parser_state| Parse_many of Base.unit parser_state| Terminate