package SZXX
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Module Expert.SSTSource
Advanced: for use with SZXX.Xml.SAX.Expert.Stream.folder
Advanced: for use with SZXX.Zip.stream_files
Extract the SST from an XLSX document. This function will stop reading from the Feed as soon as it has retrieved the SST.
Extract the SST from an XLSX file. This function does not advance the file cursor. It jumps around the file to only extract the SST while reading as few bytes as necessary.
Resolve a single reference into the Shared Strings Table.