package SZXX
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Module SZXX.XmlSource
Convenience function to access attributes by name
Convenience function to check whether a node has attribute xml:space="preserve"
type parser_options = {accept_html_boolean_attributes : bool;(*Invalid XML but valid HTML:
*)<div attr1="foo" attr2=bar>When this option istrue,attr2will be"bar"accept_unquoted_attributes : bool;
}IO-agnostic Angstrom.t parser.
It is not fully spec-compliant, it does not attempt to validate character encoding or reject all incorrect documents. It does not process references. It does not automatically unescape XML escape sequences automatically but unescape is provided to do so.
See README.md for examples on how to use it.
unescape "Fast & Furious" returns "Fast & Furious"