package SZXX

  1. Overview
  2. Docs

Module SZXX.XmlSource

Sourcetype attr_list = (string * string) list
Sourceval sexp_of_attr_list : attr_list -> Sexplib0.Sexp.t
Sourceval get_attr : attr_list -> string -> string option

Convenience function to access attributes by name

Sourcemodule DOM : sig ... end
Sourcemodule SAX : sig ... end

IO-agnostic Angstrom.t parser.

It is not fully spec-compliant, it does not attempt to validate character encoding or reject all incorrect documents. It does not process references. It does not automatically unescape XML escape sequences automatically but unescape is provided to do so.

Call this parser repeatedly to generate a sequence of SAX.node values. Those SAX.node values can then be folded using SAX.To_DOM or SAX.Stream.

See README for more information.

Sourceval unescape : string -> string

unescape "Fast & Furious" returns "Fast & Furious"