package SZXX
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Module Zip.DataSource
type 'a parser_state = | Success of 'a| Failed of {error : string;unconsumed : string;(*This string contains the start of the data not consumed by the parser. More data might have been present but it was not kept to prevent unbounded memory usage.
*)
}| Terminated_early of {unconsumed : string;(*This string contains the start of the data not consumed by the parser. More data might have been present but it was not kept to prevent unbounded memory usage.
*)
}| Incomplete
type 'a t = | Skip| Fast_skip| String of string| Bigstring of Core.Bigstring.t| Fold_string of 'a| Fold_bigstring of 'a| Parse of 'a parser_state| Parse_many of unit parser_state| Terminate